package org.cometd.oort;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.cometd.Channel;
import org.cometd.Client;
import org.cometd.Message;
import org.cometd.MessageListener;
import org.mortbay.cometd.ClientImpl;
import org.mortbay.component.AbstractLifeCycle;
import org.mortbay.util.LazyList;
import org.mortbay.util.MultiMap;
import org.mortbay.util.ajax.JSON;
import org.mortbay.util.ajax.JSON.Output;

/* ------------------------------------------------------------ */
public class Seti extends AbstractLifeCycle
{
    public final static String SETI_ATTRIBUTE="org.cometd.oort.Seti";
    public final static String SETI_SHARD="seti.shard";
    
    final String _setiId;
    final String _setiChannelId;
    final String _shardId;
    final Oort _oort;
    final Client _client;
    final ShardLocation _allShardLocation;
    final Channel _setiIdChannel;
    final Channel _setiAllChannel;
    final Channel _setiShardChannel;
    
    final ConcurrentMap<String, Location> _uid2Location = new ConcurrentHashMap<String, Location>();
    
    /* ------------------------------------------------------------ */
    public Seti(Oort oort, String shardId)
    {
        _oort=oort;
        _client = _oort.getBayeux().newClient("seti");
        _setiId=_oort.getURL().replace("://","_").replace("/","_");
        _shardId=shardId;

        _setiChannelId="/seti/"+_setiId;
        _setiIdChannel=_oort.getBayeux().getChannel(_setiChannelId,true);
        _setiIdChannel.setPersistent(true);
        _oort.observeChannel(_setiIdChannel.getId());
        _setiIdChannel.subscribe(_client);
        
        _setiAllChannel=_oort.getBayeux().getChannel("/seti/ALL",true);
        _setiAllChannel.setPersistent(true);
        _oort.observeChannel(_setiAllChannel.getId());
        _setiAllChannel.subscribe(_client);
        
        _setiShardChannel=_oort.getBayeux().getChannel("/seti/"+shardId,true);
        _setiShardChannel.setPersistent(true);
        _oort.observeChannel(_setiShardChannel.getId());
        _setiShardChannel.subscribe(_client);
        
        _allShardLocation = new ShardLocation(_setiAllChannel);
         
    }

    /* ------------------------------------------------------------ */
    protected void doStart()
        throws Exception
    {
        super.doStart();
        _client.addListener(new MessageListener()
        {
            public void deliver(Client from, Client to, Message msg)
            {
                receive(from,to,msg);
            }
        });
    }
    
    /* ------------------------------------------------------------ */
    protected void doStop()
        throws Exception
    {
        // TODO stop client
    }
    
    /* ------------------------------------------------------------ */
    public void associate(final String userId,final Client client)
    {
        System.err.println("associate "+userId+" with "+client);
        _uid2Location.put(userId,new LocalLocation(client));
        userId2Shard(userId).associate(userId);
    }

    /* ------------------------------------------------------------ */
    public void disassociate(final String userId)
    {
        System.err.println("disassociate "+userId);
        _uid2Location.remove(userId);
        userId2Shard(userId).disassociate(userId);
    }

    /* ------------------------------------------------------------ */
    public void sendMessage(final String toUser,final String toChannel,final Object message)
    {
        Location location = _uid2Location.get(toUser);
        if (location==null)
            location = userId2Shard(toUser);
        
        location.sendMessage(toUser,toChannel,message);
    }

    /* ------------------------------------------------------------ */
    public void sendMessage(final Collection<String> toUsers,final String toChannel,final Object message)
    {
        // break toUsers in to shards
        MultiMap shard2users = new MultiMap();
        for (String userId:toUsers)
        {       
            ShardLocation shard = userId2Shard(userId);
            shard2users.add(shard,userId);
        }
        
        // for each shard
        for (Map.Entry<ShardLocation,Object> entry : (Set<Map.Entry<ShardLocation,Object>>)shard2users.entrySet())
        {
            // TODO, we could look at all users in shard to see if we
            // know a setiId for each, and if so, break the user list
            // up into a message for each seti-id. BUT it is probably 
            // more efficient just to send to the entire shard (unless
            // the number of nodes in the shard is greater than the
            // number of users).
            
            ShardLocation shard = entry.getKey();
            Object lazyUsers = entry.getValue();
            
            if (LazyList.size(lazyUsers)==1)
                shard.sendMessage((String)lazyUsers,toChannel,message);
            else
                shard.sendMessage((List<String>)lazyUsers,toChannel,message);
        }
    }
    
    /* ------------------------------------------------------------ */
    protected ShardLocation userId2Shard(final String userId)
    {
        return _allShardLocation;
    }

    /* ------------------------------------------------------------ */
    protected void receive(final Client from, final Client to, final Message msg)
    {
        System.err.println("SETI "+_oort+":: "+msg);

        if (!(msg.getData() instanceof Map))
            return;
        
        Map<String,Object> data = (Map<String,Object>)msg.getData();

        final String toUid=(String)data.get("to");
        final String fromUid=(String)data.get("from");
        Object message = data.get("message");
        String on = (String)data.get("on");
        
        if (on!=null)
        {
            _uid2Location.put(fromUid,new SetiLocation("/seti/"+on));
            System.err.println(_oort+":: "+fromUid+" on "+on);
        }
        else 
        {
            String off = (String)data.get("off");
            if (off!=null)
            {
                // TODO check off value
                _uid2Location.remove(fromUid,new SetiLocation(off));
                System.err.println(_oort+":: "+fromUid+" off ");
            }
        }
        
        if (message!=null && toUid!=null)
        {
            final String toChannel=(String)data.get("channel");
            final Location location=_uid2Location.get(toUid);
            final boolean to_shard = !_setiChannelId.equals(msg.getChannel());
            
            if (location!=null)
            {
                location.receive(toUid,toChannel,message);
            }
            else
            {
                
            }
            
        }
        
    }


    /* ------------------------------------------------------------ */
    /* ------------------------------------------------------------ */
    private interface Location
    {
        public void sendMessage(String toUser,String toChannel,Object message);
        public void receive(String toUser,String toChannel,Object message);
    }
    

    /* ------------------------------------------------------------ */
    /* ------------------------------------------------------------ */
    class LocalLocation implements Location
    {
        Client _client;
        
        LocalLocation(Client client)
        {
            _client=client;
        }

        public void sendMessage(String toUser, String toChannel, Object message)
        {
            _client.deliver(Seti.this._client,toChannel,message,null);
        }

        public void receive(String toUser, String toChannel, Object message)
        {
            _client.deliver(Seti.this._client,toChannel,message,null);
        }
    }

    /* ------------------------------------------------------------ */
    /* ------------------------------------------------------------ */
    class SetiLocation implements Location
    {
        Channel _channel;

        SetiLocation(String channelId)
        {
            _channel=_oort._bayeux.getChannel(channelId,true);
        }
        
        SetiLocation(Channel channel)
        {
            _channel=channel;
        }
        
        public void sendMessage(String toUser, String toChannel, Object message)
        {
            _channel.publish(Seti.this._client,new SetiMessage(toUser,toChannel,message),null);
        }

        public void receive(String toUser, String toChannel, Object message)
        {
            
        }

        public boolean equals(Object o)
        {
            return o instanceof SetiLocation &&
            ((SetiLocation)o)._channel.equals(_channel);
        }
        
        public int hashCode()
        {
            return _channel.hashCode();
        }
    }

    /* ------------------------------------------------------------ */
    /* ------------------------------------------------------------ */
    class ShardLocation implements Location
    {
        Channel _channel;
        
        ShardLocation(String shardId)
        {
            _channel=_oort._bayeux.getChannel("/seti/"+shardId,true);
            
        }
        
        ShardLocation(Channel channel)
        {
            _channel=channel;
        }
        
        public void sendMessage(final Collection<String> toUsers, final String toChannel, final Object message)
        {
            _channel.publish(Seti.this._client,new SetiMessage(toUsers,toChannel,message),null);
        }

        public void sendMessage(String toUser, String toChannel, Object message)
        {
            _channel.publish(Seti.this._client,new SetiMessage(toUser,toChannel,message),null);
        }
        
        public void receive(String toUser, String toChannel, Object message)
        {
            
        }
        
        public void associate(final String user)
        {
            _channel.publish(Seti.this._client,new SetiPresence(user,true),null);
        }
        
        public void disassociate(final String user)
        {
            _channel.publish(Seti.this._client,new SetiPresence(user,false),null);
        }
    }

    /* ------------------------------------------------------------ */
    /* ------------------------------------------------------------ */
    class SetiMessage implements JSON.Convertible
    {
        String _toUser;
        Collection<String> _toUsers;
        String _toChannel;
        Object _message;

        SetiMessage(String toUser,String toChannel, Object message)
        {
            _toUser=toUser;
            _toChannel=toChannel;
            _message=message;
        }
        
        SetiMessage(Collection<String> toUsers,String toChannel, Object message)
        {
            _toUsers=toUsers;
            _toChannel=toChannel;
            _message=message;
        }
        
        public void fromJSON(Map object)
        {
            throw new UnsupportedOperationException();
        }

        public void toJSON(Output out)
        {
            if (_toUser!=null)
                out.add("to",_toUser);
            else if (_toUsers!=null)
                out.add("to",_toUsers);
            out.add("channel",_toChannel);
            out.add("from",_setiId);
            out.add("message",_message);
        }   
    }
    
    /* ------------------------------------------------------------ */
    /* ------------------------------------------------------------ */
    class SetiPresence implements JSON.Convertible
    {
        String _user;
        boolean _on;

        SetiPresence(String user,boolean on)
        {
            _user=user;
            _on=on;
        }
        
        public void fromJSON(Map object)
        {
            throw new UnsupportedOperationException();
        }

        public void toJSON(Output out)
        {
            out.add("from",_user);
            out.add(_on?"on":"off",_setiId);
        }
    }
    
}
